package computations;

import gui.setup.OpPanel;

import java.text.NumberFormat;
import java.util.ArrayList;

import simulator.SimulatorModel;
import datamodel.Primitive;
import datamodel.System;
import desmoj.core.simulator.Experiment;
import desmoj.core.simulator.SimTime;

public class GainCalculator {

	private System system;
	
	private int MAX_NUM_OF_SIMULATIONS=15;
	
//	The relative percentage error allowed from the Operating Point
	private static double MAX_OP_DIFF = 0.03;
	
	public static final int SIMULATION=0;
	public static final int MVA=1;

	private int calcMethod = SIMULATION;
	
	private static int RUNTIME_SIMSAM=100000;
	
	private double gainProcessing=1;
	private double gainStorage=1;
	private double gainConnectivity=1;

	public GainCalculator(System sys) {
		system = sys;
	}	

	public void calculateGainProcessing() {
		NumberFormat form=NumberFormat.getInstance();
		form.setGroupingUsed(false);
		gainProcessing= new Double(form.format(
					(double)calculateLoadAtOP()/(double)system.getBaselineLoad())).doubleValue();				
	}
	public void calculateGainStorage(){
		double baselineStorage=0;
		double scaledStorage=0;
		ArrayList<Primitive> devices=System.getAllPrimitives(system.getRootNode());
		for (Primitive device : devices) {
			baselineStorage+=device.getStorageCapacity();
			scaledStorage+=(device.getStorageCapacity()*device.getUpgradeFactorStorage())*device.getReplication();
		}
		if(baselineStorage==0 || scaledStorage==0){
			gainStorage= 1;
		}else{
			NumberFormat form=NumberFormat.getInstance();
			form.setGroupingUsed(false);
			gainStorage= new Double(form.format(scaledStorage/baselineStorage)).doubleValue();
		}
	}
	public void calculateGainConnectivity(){
		double baselineConn=0;
		double scaledConn=0;
		ArrayList<Primitive> devices=System.getAllPrimitives(system.getRootNode());
		for (Primitive device : devices) {
			baselineConn+=device.getConnectivity();
			scaledConn+=(device.getConnectivity()*device.getUpgradeFactorConnectivity())*device.getReplication();
		}
		if(baselineConn==0 || scaledConn==0){
			gainConnectivity= 1;
		}else{
			NumberFormat form=NumberFormat.getInstance();
			form.setGroupingUsed(false);
			gainConnectivity= new Double(form.format(scaledConn/baselineConn)).doubleValue();
		}
	}

	public int calculateLoadAtOP() {
		calculateServiceDemands();
		if (calcMethod == SIMULATION) {
			return doSimulation();
		} else if(calcMethod==MVA) {
			return doMVA();
		}
		else{
			//error?
			return -1;
		}
	}

	public SimulatorModel simulate(int numUsers) {
		Experiment SIMSAM_Experiment = new Experiment("SIMSAM_Experiment");
		SimulatorModel simsamModel = new SimulatorModel(null,
				"SIMSAM_Experiment", true, false, System.getAllPrimitives(system.getRootNode()));
		simsamModel.setNUM_OF_USERS(numUsers);
		simsamModel.connectToExperiment(SIMSAM_Experiment);
		SIMSAM_Experiment.setShowProgressBar(false);
		SIMSAM_Experiment.tracePeriod(new SimTime(0.0), new SimTime(simsamModel
				.getSIM_TRACE_PERIOD()));
		SIMSAM_Experiment.debugPeriod(new SimTime(0.0), new SimTime(simsamModel
				.getSIM_DEBUG_PERIOD()));
		SIMSAM_Experiment.stop(new SimTime(GainCalculator.RUNTIME_SIMSAM));
		SIMSAM_Experiment.start();
		SIMSAM_Experiment.report();
		SIMSAM_Experiment.finish();
		java.lang.System.out.println("Load: " + numUsers);
		return simsamModel;
	}

	private int doSimulation() {
		int repetitions=0;
		int index = 0;// test
		int initialUsers = 10;
		double measure = 0;
		SimulatorModel sim = simulate(initialUsers);
		if (system.getOpType() == OpPanel.OP_UTILISATION) {
			double bottleneckUtilisation = 0.001;
			int n = sim.getResources().size();
			for (int i = 0; i < n; i++) {
				double utilisation = sim.getResources().get(i).avgUsage();
				if (utilisation > bottleneckUtilisation) {
					bottleneckUtilisation = utilisation;
					index = i;// test
				}
			}
			measure = bottleneckUtilisation;
		}
		double OP = system.getOP();
		// test
		java.lang.System.out.println("Bottleneckdevice: "
				+ sim.getResources().get(index).getName());
		java.lang.System.out.println("Bottleneckdevice Utilisation: "
				+ sim.getResources().get(index).avgUsage());
		// test
		while (!isOperatingPointReached(measure, OP)&& repetitions++<=MAX_NUM_OF_SIMULATIONS) {
			int load = getAdjustedLoad(sim.getNUM_OF_USERS(), measure, OP);
			if(load==0){
				return 1;
			}
			sim = simulate(load);
			if (system.getOpType() == OpPanel.OP_UTILISATION) {
				double bottleneckUtilisation = 0.001;
				int n = sim.getResources().size();
				for (int i = 0; i < n; i++) {
					double utilisation = sim.getResources().get(i).avgUsage();
					if (utilisation > bottleneckUtilisation) {
						bottleneckUtilisation = utilisation;
						index = i;// test
					}
				}
				measure = bottleneckUtilisation;
			}
			// test
			java.lang.System.out.println("Bottleneckdevice: "
					+ sim.getResources().get(index).getName());
			java.lang.System.out.println("Bottleneckdevice Utilisation: "
					+ sim.getResources().get(index).avgUsage());
			// test
		}

		return sim.getNUM_OF_USERS();
	}

	public int doMVA() {
		// insert mva here
		return 0;
	}

	public int getAdjustedLoad(int prevLoad, double measure, double OP) {
		return (int) Math.round(prevLoad * (OP / measure));
	}

	private boolean isOperatingPointReached(double measure, double op) {
		double diff = Math.abs(op - measure) / op;
		return (diff < GainCalculator.MAX_OP_DIFF);
	}

	private void calculateServiceDemands() {
		ArrayList<Primitive> devices=System.getAllPrimitives(system.getRootNode());		
		for (Primitive device : devices) {
			device.calculateOriginalSD();
			device.calcScaleSD();
		}
	}

	public void setCalcMethod(int calcMethod) {
		this.calcMethod = calcMethod;
	}

	public void setMAX_NUM_OF_SIMULATIONS(int max_num_of_simulations) {
		MAX_NUM_OF_SIMULATIONS = max_num_of_simulations;
	}

	public double getGainConnectivity() {
		return gainConnectivity;
	}

	public double getGainProcessing() {
		return gainProcessing;
	}

	public double getGainStorage() {
		return gainStorage;
	}
}
